# -*- coding: utf-8 -*-
# This enhancement will alert at alert_time_delta after the "init" event type
# if the "resolve" event type isn't detected within that time delta.

import datetime
from elastalert.ruletypes import RuleType
from elastalert.enhancements import BaseEnhancement
from elastalert.enhancements import DropMatchException
from elastalert.elastalert import ElastAlerter
from elastalert.util import lookup_es_key
from elastalert.util import ts_to_dt, ts_now
from elasticsearch.client import Elasticsearch


class AlertTimeDeltaRule(RuleType):
    """ Aggregates matches by identifiers """
    def add_data(self, data):
        match = data[0]
        match['identifiers'] = []
        for datum in data:
            match['identifiers'].append(lookup_es_key(datum, self.rules['identifier']))
        self.add_match(match)


class AlertTimeDeltaEnhancement(BaseEnhancement):
    def __init__(self, *args):
        super(AlertTimeDeltaEnhancement, self).__init__(*args)
        # This delta should probably always be >~10s to avoid falsely alerting due es/process
        # time delay
        self.alert_time_delta = datetime.timedelta(**self.rule.get('alert_time_delta'))
        self.realert = datetime.timedelta(**self.rule.get('effective_realert'))
        self.pending_alerts = {}
        self.es = Elasticsearch(host=self.rule['writeback_host'], port=self.rule['writeback_port'])

    def process(self, match):
        # This process command will run under 3 cases:
        # 1. When the event type 1 occurs, it will add "pending_alert" to the match and upload it
        # 2. When event type 2 occurs, it will try to delete the pending alert
        # 3. When the pending alert finally fires. At this point, we do nothing

        # Don't run match on the manually added alert, let it propagate
        if 'pending_alert' in match:
            # Delete other pending alerts, to simulate realert
            for key, (_id, ts) in self.pending_alerts.items():
                if ts > ts_now() and ts - ts_now() < self.realert:
                    self.es.delete(index='elastalert_status', doc_type='elastalert', id=_id)
                    del self.pending_alerts[key]
            return

        # Identifying es field to associate initial and resolve types
        identifier = self.rule.get('identifier')
        _type = lookup_es_key(match, 'type')

        if _type == self.rule.get('init_type'):
            match['pending_alert'] = True
            # Copied from ElastAlerter.get_alert_body
            body = {
                'match_body': match,
                'rule_name': self.rule['name'],
                'alert_info': {'type': 'deferred'},
                'alert_sent': False,
                'alert_time': ts_to_dt(match['@timestamp']) + self.alert_time_delta,
                '@timestamp': ts_now()
            }
            res = self.es.create(index='elastalert_status', doc_type='elastalert', body=body)
            _id = res['_id']
            # This needs to use something from the match to map pending alerts
            self.pending_alerts[lookup_es_key(match, identifier)] = (_id, ts_to_dt(match['@timestamp']) + self.alert_time_delta)
            raise DropMatchException

        elif _type == self.rule.get('resolve_type'):
            key = lookup_es_key(match, identifier)
            if key in self.pending_alerts:
                self.es.delete(index='elastalert_status', doc_type='elastalert', id=self.pending_alerts[key][0])
                del self.pending_alerts[key]
            raise DropMatchException

