# -*- coding: utf-8 -*-
import datetime
import copy

from elasticsearch.client import Elasticsearch
from elastalert.ruletypes import RuleType
from elastalert.util import ts_to_dt
from elastalert.util import dt_to_ts


class CompareHistoricDataRule(RuleType):
    """The base class for batch report alert rule
    """
    required_options = frozenset(['monitor_fields', 'delta_time', 'delta_size'])

    def __init__(self, *args):
        super(CompareHistoricDataRule, self).__init__(*args)
        self.es = Elasticsearch(host=self.rules['es_host'], port=self.rules['es_port'])
        self.batch_report_query = {'filter': {'bool': {'must': [{'term': {'_type': 'logs'}}]}}}

    def add_data(self, data):
        query = self.get_query(data)
        if query:
            # source include take the fields from rules monitor_fields in its yaml file
            res = self.es.search(index=self.rules['index'],
                                 body=query,
                                 _source_include=self.rules['monitor_fields'],
                                 ignore_unavailable=True)

            hits = res['hits']['hits']
        for document in data:
            self.check_for_alert(hits, document)

    def get_query(self, data):
        """Return the query to be used by elastalert over the documents in the
        elasticsearch index given in the rules
        """
        raise NotImplementedError

    def check_for_alert(self, hits, document):
        """Given the hits and the current document, provide your alert rule
        """
        raise NotImplementedError


class PreviousTimeAlertRule(CompareHistoricDataRule):
    """This is an abstract class that query over the data for delta time provides in the
    the yaml file
    """

    def update_time_delta(self, time_tuple, **kwargs):
        """Utility to update time_tuple as per the previous time provided that includes
        days, hours, minutes, second
        """
        return time_tuple - datetime.timedelta(**kwargs)

    def update_months_delta(self, date, delta):
        """Utility to update time_tuple as per the previous months provided
        """
        delta = - (delta)
        m, y = (date.month + delta) % 12, date.year + ((date.month) + delta - 1) // 12
        if not m:
            m = 12
        d = min(date.day, [31, 29 if y % 4 == 0 and not y % 400 == 0 else 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31][m - 1])
        return date.replace(day=d, month=m, year=y)

    def update_years_delta(self, time_tuple, years):
        """Utility to update time_tuple as per the previous years provided, years could positive or negative value
        """
        years = -(years)
        time_list = list(time_tuple)
        time_list[0] -= years

        return datetime.datetime(*time_list)

    def get_delta_time(self, timestamp, delta_time_dict):
        """Utility that uses update years, months and time delta utilities to update the time_tuple
        """
        time_tuple = copy.copy(timestamp)

        if delta_time_dict.get('years'):
            time_tuple = self.update_years_delta(time_tuple, delta_time_dict['years'])
            delta_time_dict.pop('years')

        if delta_time_dict.get('months'):
            time_tuple = self.update_months_delta(time_tuple, delta_time_dict['months'])
            delta_time_dict.pop('months')

        if delta_time_dict:
            time_tuple = self.update_time_delta(time_tuple, **delta_time_dict)

        return time_tuple

    def get_query(self, document):
        """Return query that consider document found last time
        """
        timestamp = ts_to_dt(document[0]['@timestamp'])

        # find delta timestamp for the given delta_time in rules
        previous_time = self.get_delta_time(timestamp, self.rules['delta_time'])

        if previous_time:
            # define the delta size range to consider for query
            delta_size = self.rules['delta_size']
            timestamp_from = previous_time - datetime.timedelta(**delta_size)
            timestamp_to = previous_time + datetime.timedelta(**delta_size)

            query = copy.copy(self.batch_report_query)
            query['filter']['bool']['must'].append({'range': {'@timestamp': {'from': dt_to_ts(timestamp_from),
                                                                             'to': dt_to_ts(timestamp_to)}}})
        else:
            # Do not form any query if the previous time not found
            query = None

        return query

    def check_for_alert(self, hits, document):
        """Given the hits and current document, check for alert
        """
        raise NotImplementedError


class AccountTransactionReportDailyAlertRule(PreviousTimeAlertRule):
    """This is an elastalert rule specific for logs generated by tron
    ad_daily transaction report daily run
    """

    def check_for_alert(self, hits, document):
        """Define the ATR report alert rule"""
        if len(hits):
            last_time_data = hits[0]['_source']['number_of_transactions']
        else:
            last_time_data = 0

        # A factor of .1 with 'down' means "alert if number is less than 90% of before"
        # whereas a factor of .1 with 'up' means "alert if number more than 110% of before"
        factor = self.rules.get('factor', 0)
        direction = self.rules.get('direction', 'down')

        if document.get('number_of_transactions') < last_time_data * (1 - factor) and direction in ['down', 'both']:
            document['previous_num_transactions'] = last_time_data
            self.add_match(document)
        elif document.get('number_of_transactions') > last_time_data * (1 + factor) and direction in ['up', 'both']:
            document['previous_num_transactions'] = last_time_data
            self.add_match(document)
