# -*- coding: utf-8 -*-
import datetime

import pytz
import elastalert.util as ea_util
from elastalert.ruletypes import RuleType
from elasticsearch.client import Elasticsearch

from modules.time_utils import previous_occurence_of
from modules.time_utils import time_from_string
from modules.time_utils import LATER
from modules.time_utils import EARLIER


def _assert_string(str_):
    if not isinstance(str_, basestring):
        raise ValueError("{0!r} doesn't look like a basestring".format(str_))

    return str_


def _dict_to_timedelta(dict_):
    return datetime.timedelta(**dict_)


def _maybe_transform(dict_, **transforms):
    for key, transformer in transforms.iteritems():
        try:
            val = dict_[key]
        except KeyError:
            pass
        else:
            dict_[key] = transformer(val)


class NotEnoughData(Exception):
    # Used by DailyBatchWatchdog.calculate_runtime_mean when there's no data in
    # ES to make an average out of.
    pass


class DailyBatchWatchdog(RuleType):
    """Watch for a daily batch's starts, terminations, and run times.

    So, it's pretty common that somebody has to keep an eye on batches that run
    daily, and raise issues on certain conditions. This has become so
    mechanical we can actually teach our robots to do this for us. Here are the
    things we currently care about:
        - The batch took took less than minimum:
            This is to be alerted of weird conditions that might make the batch
            refuse to work, but still terminate.
        - The batch took too more than maximum:
            The batch has took an unreasonable amount of time.

        - The batch took less time than normal:
        - The batch took more time than normal:
            Upon completion of the batch, we look at the time it took to
            complete. If it deviated too much from the previous times, we
            probably took a performance hit we weren't expecting or something
            changed in non-trivial ways, and we should investigate to be sure
            everything is alright.
            Neither of these condition are triggered if the batch took less
            less than minimum or it took too more than maximum.

        - The batch didn't start in time:
            This normally happens when a batch depends on some upstream job to
            complete in order for Tron to start it. If this happens, we should
            take a look at the state of things.
        - The batch didn't terminate in time:
            This normally happens when a batch encounters something that
            prevented it to complete. The OOM killer was, at the time of
            writing, the most common culprit for things to "just disappear".
    """

    # Just a bunch of constants to symbolize the issues we look for.
    # Also used to provide a human readable description.
    TOOK_LESS_THAN_MINIMUM = 'took less than minimum'
    TOOK_MORE_THAN_MAXIMUM = 'took more than maximum'
    TOOK_LESS_TIME_THAN_NORMAL = 'took less time than normal'
    TOOK_MORE_TIME_THAN_NORMAL = 'took more time than normal'
    FAILED_TO_START_IN_TIME = 'failed to start in time'
    FAILED_TO_FINISH_IN_TIME = 'failed to finish in time'

    required_options = frozenset([
        'batch_name',
    ])

    def __init__(self, rules, *args, **kwargs):
        # Pre-processed in load_options.
        # We should setdefault with the final object type.
        rules.setdefault('buffer_time', datetime.timedelta(hours=1))
        rules.setdefault('alert_subject', '%s on %s: %s')
        rules.setdefault('alert_subject_args', [
            'batch_name',
            'execution_day',
            'problem_description',
        ])

        # Our custom stuff. Not pre-processed.
        # We should setdefault with raw types (as it would be in the YAML).
        rules.setdefault('runtime_mean_lookback', dict(days=7))

        super(DailyBatchWatchdog, self).__init__(rules, *args, **kwargs)

        # Transform stuff from raw to final types on present information.
        _maybe_transform(
            self.rules,
            batch_name=_assert_string,
            scheduled_start_time=time_from_string,
            latest_tolerable_start_time=time_from_string,
            latest_tolerable_end_time=time_from_string,
            minimum_tolerable_runtime=_dict_to_timedelta,
            maximum_tolerable_runtime=_dict_to_timedelta,
            runtime_mean_lookback=_dict_to_timedelta,
            tolerable_runtime_deviation_from_mean=float,
        )

        # At the time of writing, garbage_collect needs to be able to generate
        # two distinct matches at the same time that should not be grouped
        # together. If we figure out a way to tell EA to not try to group these
        # together, we can drop this restriction.
        realert = self.rules.get('realert')
        if realert is not None and realert != datetime.timedelta(seconds=60):
            raise ea_util.EAException(
                "This rule wasn't designed to have a customizable realert.",
            )
        self.rules['realert'] = False

        self.rules['original_filter'] = self.rules['filter']
        # This rule type only cares about the configured batch.
        self.rules['filter'] = self.rules['filter'] + [{'term': {
            'logger_name': self.rules['batch_name'],
        }}]

        self.rules['original_include'] = self.rules['include']
        # We need these fields to emit alerts with the right descriptiions.
        self.rules['include'] = self.rules['include'] + [
            'state',
            'runtime',
            'logger_name',
        ]

        scheduled_times_keys = (
            'scheduled_start_time',
            'latest_tolerable_start_time',
            'latest_tolerable_end_time',
        )
        if (
            any(f in self.rules for f in scheduled_times_keys) and
            not all(f in self.rules for f in scheduled_times_keys)
        ):
            raise ea_util.EAException(
                "If you define any of the fields '{0}', you need to define "
                "all of them."
                .format(
                    ', '.join(scheduled_times_keys),
                ),
            )

        self.es = Elasticsearch(
            host=self.rules['es_host'],
            port=self.rules['es_port'],
        )

    def _fire_alert(self, reason, detection_dt):
        self.add_match({
            self.rules['timestamp_field']: detection_dt,
            'batch_name': self.rules['batch_name'],
            'execution_day': str(detection_dt.date()),
            'problem_description': reason,
        })

    def add_data(self, data):
        """React to batch terminating and check if they took the expected
        amount of time."""
        for datum in data:
            # We only care about how long things took, here.
            if datum.get('state') != 'runtime':
                continue

            # Extract all the values.
            runtime = int(datum['runtime'])
            when = datum[self.rules['timestamp_field']]
            minimum = ea_util.seconds(self.rules.get(
                'minimum_tolerable_runtime',
                datetime.timedelta.min,
            ))
            maximum = ea_util.seconds(self.rules.get(
                'maximum_tolerable_runtime',
                datetime.timedelta.max,
            ))
            tolerable_dev = self.rules.get(
                'tolerable_runtime_deviation_from_mean',
                float('inf'),
            )
            try:
                mean = self.calculate_runtime_mean(
                    when,
                    self.rules['runtime_mean_lookback'],
                    self.rules['filter'],
                )
            except NotEnoughData:
                mean = None

            # Decide things.
            if runtime < minimum:
                self._fire_alert(self.TOOK_LESS_THAN_MINIMUM, when)
            elif mean is not None and runtime < mean * (1 - tolerable_dev):
                self._fire_alert(self.TOOK_LESS_TIME_THAN_NORMAL, when)

            if runtime > maximum:
                self._fire_alert(self.TOOK_MORE_THAN_MAXIMUM, when)
            elif mean is not None and runtime > mean * (1 + tolerable_dev):
                self._fire_alert(self.TOOK_MORE_TIME_THAN_NORMAL, when)

    def _query_index(self, start_dt, end_dt):
        return ea_util.format_index(self.rules['index'], start_dt, end_dt)

    def _query_body(self, state, start_dt, end_dt, filters):
        return {'filter': {'bool': {'must': [
            {'range': {self.rules['timestamp_field']: {
                'gte': ea_util.dt_to_ts(start_dt),
                'lte': ea_util.dt_to_ts(end_dt),
            }}},
            {'term': {
                'state': state,
            }},
        ] + filters}}}

    def garbage_collect(self, end_dt):
        """Check if things have started and/or terminated when we expect them
        to.

        This could be done by accumulating local state in add_data, but I'd
        rather query ES for past events when I know a particular point in time
        has passed in order to keep as little local state as possible. Makes
        reasoning about EA's restarts a non-issue, and less state is always
        less things we have to care about.
        """
        if 'scheduled_start_time' not in self.rules:
            return

        # Make it a pytz tzinfo
        end_dt = pytz.utc.normalize(end_dt)

        # UTC is transitionless. No need to normalize.
        start_dt = end_dt - self.rules['buffer_time']

        def _check_time_has_past_and_alert_if(
            time_field,
            decider,
            alert_name,
        ):
            if time_field not in self.rules:
                return False

            latest_time_occurence = previous_occurence_of(
                end_dt,
                self.rules[time_field],
                LATER,
            )

            if not (start_dt <= latest_time_occurence < end_dt):
                return False

            if not decider(latest_time_occurence):
                return False

            self._fire_alert(alert_name, end_dt)

        scheduled_dt = previous_occurence_of(
            end_dt,
            self.rules['scheduled_start_time'],
            EARLIER,
        )

        def count_events(state, target_dt):
            return self.es.count(
                index=self._query_index(scheduled_dt, target_dt),
                body={'query': {'filtered': self._query_body(
                    state,
                    scheduled_dt,
                    target_dt,
                    self.rules['filter'],
                )}},
                ignore_unavailable=True,
            )['count']

        _check_time_has_past_and_alert_if(
            'latest_tolerable_start_time',
            lambda target_dt: count_events('startup', target_dt) < 1,
            self.FAILED_TO_START_IN_TIME,
        )

        _check_time_has_past_and_alert_if(
            'latest_tolerable_end_time',
            lambda target_dt: (
                count_events('startup', target_dt) -
                count_events('shutdown', target_dt)
            ) > 0,
            self.FAILED_TO_FINISH_IN_TIME,
        )

    def calculate_runtime_mean(self, end_dt, lookback, filters):
        start_dt = end_dt - lookback
        result = self.es.search(
            _source=False,
            fields=['runtime'],
            index=self._query_index(start_dt, end_dt),
            body=self._query_body(
                'runtime',
                start_dt,
                end_dt,
                self.rules['filter'],
            ),
            ignore_unavailable=True,
        )

        if result['hits']['total'] < 1:
            raise NotEnoughData()

        # Some old data has strings with numbers, instead of just numbers.
        # After we fix that on prod data, let's start using actual aggregations
        values = [
            int(hit['fields']['runtime'][0])
            for hit in result['hits']['hits']
        ] or [0]
        return sum(values) / len(values)
