# -*- coding: utf-8 -*-
from elastalert.alerts import EmailAlerter, Alerter, BasicMatchString
import datetime
import copy

from elastalert.util import lookup_es_key, ts_to_dt, ts_now, elastalert_logger
from elasticsearch.client import Elasticsearch
from elastalert.config import alerts_mapping, load_alerts


class SelfServiceAlerter(Alerter):

    required_options = set(['escalate_alerts', 'alert_escalation_time', 'username_field', 'alert_text', 'writeback_host', 'writeback_port'])

    def __init__(self, *args):
        super(SelfServiceAlerter, self).__init__(*args)
        # This delta should probably always be >~10s to avoid falsely alerting due es/process
        # time delay
        self.alert_time_delta = datetime.timedelta(**self.rule.get('alert_escalation_time'))
        self.es = Elasticsearch(host=self.rule['writeback_host'], port=self.rule['writeback_port'])
        self.es_index = self.rule.get('writeback_index', 'elastalert_status')
        # Is updated later in the alert function

        self.initial_alert_config = copy.copy(self.rule)
        self.escalate_alert_config = copy.copy(self.rule)
        self.initial_alert_config['email'] = ""  # Need an email field for the alert to instantiate. It's updated later.
        self.initial_alerter = EmailAlerter(self.initial_alert_config)
        self.rule['escalate_alerts'] = load_alerts(self.rule, alert_field=self.rule['escalate_alerts'])

    def alert(self, matches):
        # CHECK FOR PENDING ALERTS HERE, if it is pending, all we need to do is call the escalate alert
        for match in matches:
            actor = lookup_es_key(match, self.rule.get('username_field'))

            # If the alert is not already pending, it will be set to pending
            if 'pending_alert' not in match and actor:
                if actor.endswith("@yelp.com"):
                    self.initial_alert_config['email'] = [actor]
                else:
                    self.initial_alert_config['email'] = [actor + '@yelp.com']
                # Since the same alerts use the same instance, we need to check if this has already been added
                if "https://selfservealerts.admin.yelp.com" not in self.initial_alert_config['alert_text']:
                    self.initial_alert_config['alert_text'] += \
                        "\n\nPlease address this alert by visiting https://selfservealerts.admin.yelp.com"
                self.initial_alerter.alert([match])
                match_string = BasicMatchString(self.initial_alert_config, match)
                match_string.text = ''
                match_string._add_custom_alert_text()
                self.set_pending_alert(match, actor, match_string.text)
            # If the alert was already pending when this method was hit, it is time to trigger it
            else:
                for escalate_alert in self.rule['escalate_alerts']:
                    escalate_alert.alert([match])
                self.upload_past_alert(match, actor)

    # -*- coding: utf-8 -*-
    # This enhancement will alert at alert_time_delta after the "init" event type
    # if the "resolve" event type isn't detected within that time delta.
    def upload_past_alert(self, match, actor):
        body = {
            '@timestamp': match.get(self.rule.get('timestamp_field', '@timestamp'), ts_now()),
            'match_body': match,
            'rule_name': self.rule['name'],
            'alert_info': {'type': 'deferred'},
            'actor': actor,
            'pending_alert': True,
            'escalated': True,
        }

        try:
            self.es.create(index=self.es_index, doc_type='past_elastalert', body=body)
        except Exception:
            # TODO: identification of alert
            elastalert_logger.exception('Error uploading past pending alert to Elasticsearch')

    def set_pending_alert(self, match, actor, text):
        match['pending_alert'] = True
        match['actor'] = actor
        match['text'] = text
        # Copied from ElastAlerter.get_alert_body

        # TODO: make match_body indexable so no need to repeat actor and pending_alert
        body = {
            'match_body': match,
            'rule_name': self.rule['name'],
            'alert_info': {'type': 'deferred'},
            'alert_sent': False,
            'alert_time': ts_to_dt(match.get(self.rule.get('timestamp_field', '@timestamp'))) + self.alert_time_delta,
            '@timestamp': ts_now(),
            'actor': actor,
            'pending_alert': True,
            'text': text,
            'alerts': [],
            'require_justification': self.rule.get('require_justification', False)
        }

        for escalate_alert in self.escalate_alert_config.get('escalate_alerts'):
            for (alert_name, alert_config) in escalate_alert.items():
                if alert_name in alerts_mapping and alert_name not in body['alerts']:
                    body['alerts'].append(alert_name)
        try:
            self.es.create(index=self.es_index, doc_type='elastalert', body=body)
        except Exception:
            elastalert_logger.exception('Error uploading pending alert to Elasticsearch')
