# -*- coding: utf-8 -*-
"""
Rule types that count one document multiple times according to a field in the
document. This rule may only be used when the contents of the document may be
discarded (i.e. use_count_query would have sufficed), as it does not
preserve them.
"""
import copy
import logging

from elastalert.ruletypes import FrequencyRule
from elastalert.ruletypes import SpikeRule
from elastalert.ruletypes import FlatlineRule
from elastalert.ruletypes import EventWindow
from elastalert.util import lookup_es_key


def verify_integer_field(document, rule, count_field, allow_zero=False):
    count = lookup_es_key(document, count_field)
    # Attempt to convert strings to ints
    if isinstance(count, basestring):
        try:
            count = int(count)
        except ValueError:
            logging.warning(
                'Field %s is a string which could not be converted'
                'into an integer for rule %s' % (count_field, rule['name']))
            return None
    if count is None:
        logging.warning(
            'Did not find field %s representing document count in '
            'document for rule %s' % (count_field, rule['name']))
        return None
    if isinstance(count, float):
        count = int(count)
    if not isinstance(count, (int, long)):
        logging.warning(
            'Non-integer value of field %s representing document '
            'count in document for rule %s: %s'
            % (count_field, rule['name'], count))
        return None
    if count < 0 or (count == 0 and not allow_zero):
        logging.warning(
            'Negative or zero value of %s representing document '
            'count in document for rule %s: %d'
            % (count_field, rule['name'], count))
        return None
    return count


class SumOfFieldRuleFactory(type):
    """
    Metaclass that creates a subclass of the class's backing_rule_type.
    backing_rule_type must implement add_count_data.
    """
    # This allows the add_data to use super(...).add_count_data(...) to call
    # the add_count_data method of the appropriate backing RuleType, without
    # having to reimplement all the common logic for each backing_rule_type.

    def __new__(mcs, name, _, dct):

        def add_data(self, data):
            for document in data:
                count = verify_integer_field(
                    document, self.rules, self.rules['count_field'])
                ts = lookup_es_key(document, self.ts_field)
                if count and ts:
                    super(self.__class__, self).add_count_data({ts: count})

        backing_rule_type_cls = dct['backing_rule_type']
        dct['required_options'] = (
            dct['backing_rule_type'].required_options | frozenset(['count_field']))
        dct['add_data'] = add_data
        return type.__new__(mcs, name, (backing_rule_type_cls,), dct)


class SumOfFieldFrequencyRule(object):
    __metaclass__ = SumOfFieldRuleFactory
    backing_rule_type = FrequencyRule


class SumOfFieldSpikeRule(object):
    __metaclass__ = SumOfFieldRuleFactory
    backing_rule_type = SpikeRule


class MultipleSumOfFieldSpikeRule(SpikeRule):
    """ A rule type for matching docs by counting fields. Helps avoid
    writing multiple yaml files, treats a single file as containing N rule
    definitions.
    """

    required_options = set(['count_fields'])

    def add_data(self, data):
        for document in data:
            ts = lookup_es_key(document, self.ts_field)
            for field in self.rules['count_fields']:
                count = verify_integer_field(
                    document, self.rules, field)
                if count and ts:
                    # Set matching_field for add_match
                    self.matching_field = field
                    self.handle_event(document, count, field)

    def add_match(self, event, qk):
        doc_copy = copy.deepcopy(event)
        # Add in a field that will be used as query key, to ensure
        # each individual field can alert separately
        doc_copy['matching_field'] = self.matching_field
        super(MultipleSumOfFieldSpikeRule, self).add_match(doc_copy, qk)


class MultipleSumOfFieldFlatlineRule(FlatlineRule):
    """ A rule type for matching docs by counting fields. Helps avoid
    writing multiple yaml files, treats a single file as containing N rule
    definitions.
    """

    required_options = set(['count_fields'])

    def __init__(self, *args, **kwargs):
        super(MultipleSumOfFieldFlatlinerule, self).__init__(*args, **kwargs)
        import cProfile, pstats, StringIO
        self.pr = cProfile.Profile()
        self.x = 0


    def add_data(self, data):
        x += 1
        pr.enable()
        for document in data:
            ts = lookup_es_key(document, self.ts_field)
            for field in self.rules['count_fields']:
                count = verify_integer_field(
                    document, self.rules, field, allow_zero=True)
                if count is not None and ts:
                    field_events = EventWindow(self.rules['timeframe'], getTimestamp=self.get_ts)
                    self.occurrences.setdefault(field, field_events).append((document, count))
        for field in self.occurrences.keys():
            # Set matching_field for add_match to use
            self.matching_field = field
            self.check_for_match(field)
        pr.disable()
        print x
        if x == 15:
            s = StringIO.StringIO()
            ps = pstats.Stats(pr, stream=s).sort_stats('cumulative')
            ps.print_stats()
            print s.getvalue()
            import pdb;pdb.set_trace()
            pass

    def garbage_collect(self, ts):
        # Don't alarm if there are no events
        pass

    def add_match(self, event):
        doc_copy = copy.deepcopy(event)
        # Add in a field that will be used as query key, to ensure
        # each individual field can alert separately
        doc_copy['matching_field'] = self.matching_field
        super(MultipleSumOfFieldFlatlineRule, self).add_match(doc_copy)
