# -*- coding: utf-8 -*-
"""
There's a few things you must be *really* aware of, and have well present, when
dealing with non-trivial operations on aware datetimes and times.

Vocabulary
==========

- Universal time: this is mostly UTC (or TAI, if you're a super stickler).
They're meant to be a way of counting time forward, no more. UTC wants to
measure Earth's laps around itself and have each take as close to 86400 seconds
as possible (hence the leap seconds). TAI is atomically defined (bla bla atoms
vibrating bla bla), and wants to count the number of times an atomic SI second
occurs (relativity problems apply, but those have been dutifully corrected,
AFAICT), and it's the basis for defining UTC. Everybody uses UTC as their base
time. Mostly everybody is oblivious to leap seconds.

- UTC offsets (UO): PDT, PST, WEST, WET, etc. These are *really* just offsets
from UTC. You can totally express any datetime in them, but that might be
confusing to people unless you also tell then which UTC offset you're using
(and they know what to do with that information).

- Timezone (TZ): a non trivially defined frame of reference that humans use to
talk about time. It typically defines an UTC offset as the "expressing base"
for time during a span of time (e.g., from spring to fall we talk about time
in PDT, and we use PST the rest of the year). As an example, check
https://pb.yelpcorp.com/143085 for what LA uses. These are defined by law, most
commonly.

- Local time: The numbers you see the hands pointing to on a wall clock. It's
almost sufficient, expect when we have to deal with UO transitions on a TZ. So,
It's insufficient to specify a point in time.

- Timezone discontinuities and overlaps: As we've noted before, TZs change the
UO they use. The most appropriate example of such changes is DST. Humans have
strong opinions about DST, and most northern hemisphere timezones changes their
UO when spring and autumn begin (or some dates close to that), purportedly to
make better use of the available sunlight time. There are countries in the
southern hemisphere that also observe DST, and they normally have the direction
in their jumps reversed. Considering DST shifts is fine and dandy if you're
always working with values that have their UO coupled to them. However, it
becomes troublesome for local times. Here are the two most common transitions
(from a northern hemisphere perspective):

    - Spring forward: This happens around when spring begins. We shift an hour
    forward. This means that we'll see the clocks on our computers displaying
    some sequence of values like:
        - 1:50
        - 1:59
        - 3:00
        - 3:01
    This is a discontinuity in the local way of representing those moments in
    time. If someone says they did something at 2:10, a normal human will say
    "but that time never happened!", and rightly so. That's a non-sensical
    representation.

    - Fall back: This happens around when autumn begins. We shift an hour
    back. This means that we'll see the clocks on our computers displaying
    some sequence of values like:
        - 2:50
        - 2:59
        - 2:00
        - 2:01
    This is a overlap in the local way of representing those moments in time.
    If someone says they did something at 2:10, a normal human will say
    "oh wait, which one of the 2:00 that happened?", and rightly so. Therefore,
    During those two hours (the first and second time it was the second hour)
    it's insufficient to tell only the hand position, but you also need to say
    if it was the second hour "before the jump" or "after the jump".

UO-aware vs. TZ-aware
=====================

There's two terms I'm going to use along this module that you must pay
particular attention to: UO-aware and TZ-aware. I'll apply these to non-naive
datetime and time objects.

If a time/datetime object is UO-aware, that means its `tzinfo` represents a
UTC offset. It's a precise definition of the number of seconds that have
elapsed since the Epoch of your choice. Normally, a tzinfo object knows which TZ
it represents (remember that a TZ uses various UOs along time). pytz's
timezones certainly do.

An UO-aware datetime object specifies very clearly (short of leap seconds) a
point in time.

An UO-aware time object specifies the number of seconds since an UTC beginning
of day.

If a time/datetime object is TZ-aware, that means its `tzinfo` represents a
Timezone. However, we should disregard any attempt it makes at defining an UTC
offset. It's quite unfortunate we have to have the discipline to remember and
trace the `tzinfo` and its level of authority (TZ only as opposed to UO and
TZ), but pytz made the sad decision to not distinguish properly between both.
When you run `pytz.timezone('Some/Place')`, that will return *some* UTF offset
that belongs to the Timezone 'Some/Place'. Some versions of pytz return the
first UO in that TZ, some other versions return... whatever they feel like.
In any case, short of having different classes to represent UOs and TZs, pytz
is probably making a very decent tradeoff. Given a Timezone, you can only know
which UO you should use if you have a point in time. What's the Timezone of
California nowadays? It depends if you're asking me that during spring/summer
or autumn/winter. Also, things were not always like this. We *have* changed the
dates we go to and from DST.

A TZ-aware datetime object probably specifies a poit in time. It might not
during transition times. Try and to run
`dt.tzinfo.localize(dt.replace(tzinfo=Nonea, is_dst=None))`. If it doesn't
raise anything, that dt does define a point in time. Otherwise, it defines
either zero (during spring forward) or more (two, for example, during fall
back).

A TZ-aware time object specifies the position of a wall clock's arms in
somewhere that observes its Timezone. It's impossible to tell what's the offset
from the last UTC midnight without combining it with a date. Even so, you fall
into the same problems as the previous paragraph.

Timezone conversions
====================

One piece of advice I leave you here is I've found it very useful to reduce my
available operations to only consider two operations with timezones:

    - Normalization: Given an UO-aware datetime object, if you want the
    representation of it in another timezone, do `target_tz.normalize(uo_dt)`.
    This will basically figure out the origin dt as UTC, walk the UOs in that
    timezone until the moment that dt represents, and return a new datetime
    in that UO (properly adjusted).

    - Localization: Given a naive datetime object, figure out what UO-aware
    datetime object represents it in that timezone. This is subject to the
    discontinuities UO transitions in that timezone might beget. Always specify
    the `is_dst` argument. If you don't, you're telling pytz it's ok to take a
    guess on those discontinuities. That's very irresponsible, since it'll
    make your reasoning about it indeterminate (unless you know).

So, what about `astimezone`? `normalize`'s docs say they serve different
purposes, but they really don't. Reading the sources [1] and [2], you can see
that `astimezone`  and `normalize` are basically the same: they just bring the
origin value to UTC and ask the target timezone to return its representation of
it. `astimezone` has the added value of being implemented in C. If you think
about it, that's the only way to transfer between TZs: given an UTC
representation, figure out what is the UO that that TZ is using at that time,
and return a new object with that UO (and the appropriate offset math for them
to represent the same instant). Normalization and timezone conversion are two
names for the same things, they just vary on the focus of the concept.

So, in order to makes things simpler, I ended up using only the tzinfo objects
to represent anything that's timezone related. This meant leaving `astimezone`,
which helped me a lot on reasoning about this, because:
    - it's the right question: "How does this timezone represent this instant?"
    - one less unnecessary (and potentially confusing) concept

[1]: https://hg.python.org/cpython/file/ca78b9449e04/Modules/datetimemodule.c#l4412
[2]: http://bazaar.launchpad.net/~stub/pytz/devel/view/361/src/pytz/tzinfo.py#L189
"""

import re
import datetime

import pytz
from dateutil.parser import parse as dateutil_parser
from tzlocal import get_localzone


# The purpose of this regex is to provide a filter for the input we accept to
# parse. The actual parsing is done by dateutil_parser.
TIME_STRING_FORMAT_REGEX = re.compile(
    r"""
        ^
        (?P<time>
            (?P<hours>\d{1,2})          # Mandatory hours '12'
            :\d{2}                      # Mandatory minutes ':30'
            (?::\d{2})?                 # Maybe seconds ':59'
            (?P<period>\s*[aApP][mM])?  # Maybe period 'AM' or 'PM'
        )
        (?:
            \s+
            (?P<timezone_name>[%s]+) # Maybe a timezone name.
        )?
        $
    """ % (
        re.escape(''.join(set(''.join(pytz.all_timezones))))
    ),
    re.VERBOSE,
)


def time_from_string(str_):
    """Interpret a string as defining a TZ-aware time object.

    Mostly a wrapper around dateutil_parser to make sure we only accept time
    definitions. Uses `pytz.timezone` for the TZ.

    Here are some examples of the intended input:
        - '3:00 PM America/Los_Angeles'
        - '2:00 AM'
        - '13:00'
        - '13:00:02'

    When the timezone is not defined, we use the environment's definition of
    local TZ. All forms accept a timezone name at the end.

    :param str_ basestring: The text to be interpreted.
    :rtype: A TZ-aware datetime.time.
    """
    match = TIME_STRING_FORMAT_REGEX.match(str_.strip())
    if not match:
        raise ValueError(
            "Given string ({0!r}) doesn't conform to the regex "
            "TIME_STRING_FORMAT_REGEX".format(
                str_,
            ),
        )
    parts = match.groupdict()

    period = parts['period'] or ''
    if int(parts['hours']) > 12 and period.strip():
        raise ValueError(
            "Given string ({0!r}) has hours bigger than 12 but also defines "
            "a period. This is confusing, and I'm refusing to parse it."
            .format(str_)
        )

    timezone_name = parts['timezone_name']
    if timezone_name:
        tzinfo = pytz.timezone(timezone_name)
    else:
        tzinfo = get_localzone()

    return dateutil_parser(parts['time']).time().replace(tzinfo=tzinfo)


EARLIER = 'EARLIER'
LATER = 'LATER'
DST = 'DST'
STD = 'STD'
RAISE_EXCEPTION = 'RAISE_EXCEPTION'
TZ_AMBIGUITY_RESOLUTION_STRATEGY = set([
    EARLIER,
    LATER,
    DST,
    STD,
    RAISE_EXCEPTION,
])


def _resolve_tz_ambiguity(tzinfo, dt, strategy):
    """A wrapper around pytz.tzinfo.BaseTzInfo.localize that lets you ask for
    the earlier or later value from the interpretation possibilities (besides
    STD and DST).

    :param pytz.tzinfo.BaseTzInfo tzinfo: Symbolizes the TZ where `dt` should
        be interpreted by.
    :param datetime.datetime dt: The naive datetime object we want lo localize.
    :param string strategy: The strategy to use when we need to resolve
        ambiguous `dt`s. RAISE_EXCEPTION will just raise whatever
        BaseTzInfo.localize did. All the available strategies should be in
        TZ_AMBIGUITY_RESOLUTION_STRATEGY.

    :rtype: A UO-aware datetime.datetime object.

    Note: if we take the definition of TZ to the letter, there could be more
    than two possible values. However, this is not supported by pytz, nor by
    this method.
    """
    if dt.tzinfo is not None:
        raise ValueError(
            "dt has a declared timezone, and that's confusing. Refusing to "
            "operate on it",
        )
    if strategy not in TZ_AMBIGUITY_RESOLUTION_STRATEGY:
        raise ValueError(
            'strategy must be a value present in '
            'TZ_AMBIGUITY_RESOLUTION_STRATEGY'
        )

    try:
        return tzinfo.localize(dt, is_dst=None)

    except pytz.InvalidTimeError as e:
        if strategy is RAISE_EXCEPTION:
            raise e  # If you don't want this to raise, choose another strategy

        dst = tzinfo.localize(dt, is_dst=True)
        std = tzinfo.localize(dt, is_dst=False)
        earlier, later = sorted([dst, std])

        try:
            return tzinfo.normalize({
                EARLIER: earlier,
                LATER: later,
                DST: dst,
                STD: std,
            }[strategy])
        except KeyError:
            raise Exception(
                "The given strategy is not being contemplated in the current"
                "implementation."
            )


def _force_time_into_datetime(dt, t, tz_ambiguity_resolution):
    """Return the UO-aware datetime representation of when `t` happens in
    `t.tzinfo`'s interpretation of `dt`'s day.

    It wasn't thought at all about the calendar this uses. TZ- and UO-ness was
    thoroughly thought out.

    :param dt datetime.datetime: A UO-aware datetime that represents the target
        day where we want `t` represented at.
    :param t datetime.time: The TZ-aware time we want represented at `dt`'s
        day'. We will first normalize `dt` as seen in `t.tzinfo`.
    :param tz_ambiguity_resolution: How to disambiguate the resulting time.
    :rtype: A UO-aware datetime object normalized as observable in `dt.tzinfo`.
    """
    if dt.tzinfo is None:
        raise ValueError(
            "dt doesn't have a declared UTC offset. Refusing to operate on it",
        )
    if t.tzinfo is None:
        raise ValueError(
            "t doesn't have a declared timezone. Refusing to operate on it",
        )

    return dt.tzinfo.normalize(
        _resolve_tz_ambiguity(
            t.tzinfo,
            datetime.datetime.combine(
                t.tzinfo.normalize(dt),
                t.replace(tzinfo=None),
            ),
            tz_ambiguity_resolution,
        ),
    )


PYTZ_BASE_CLASSES = (pytz.tzinfo.BaseTzInfo, type(pytz.utc),)


def previous_occurence_of(dt, t, tz_ambiguity_resolution):
    """Find out the last time `t` happened before `dt`, or an approximation if
    `t` didn't happen in the 24 hours before `dt`.

    This method will go to great lengths to figure out some "approximation"
    when dealing with TZ transitional periods, as opposed to looking further
    back to the previous day where that would be a valid representation.

    Let's build up an exmaple. US/Eastern obverves DST. In 2002, it "springs
    forward" on the 7th of April, going from EST to EDT, skipping the entire
    hour of 2 AM. Let say we wanted to find out when was the last time 2:30 AM
    happened before 3 AM on that day. A human would just do like "Uhhh, it
    really didn't", maybe ask for details, etc. This function can't really
    dialog with you, so we make you tell it what exact interpretation of
    2:30 AM you want. A variation of this problem also happens when we "fall
    back" one hour. In this case, instead of having 0 instants that have that
    representation, we have two. So, we also have to make a decision.

    The "fall back" scenario is easy to understand: we have to pick one of
    those occurences. I want to talk a bit more about the "spring forward" one,
    which seems to lack any kind of sense.

    Timezones define which UOs humans use to identify an instant. In this case,
    we switch from using EST to use EDT at 2:00 AM EST. I could also say that
    time as 3:00 AM EDT. In fact, I could use either EST or EDT exclusively
    to express instants. It would be perfectly valid, but confusing for the
    people I was talking to. They expect me to express myself according to our
    TZ, not using a random UO just because I like it.

    With that in mind, even though no TZ compliant representation of time ever
    was 2:30 AM, both the UOs used that day in that TZ have seen 2:30. If you
    think of it, 2:30 EST is 3:30 US/Eastern, and 2:30 EDT is 1:30 US/Eastern.
    These are the two instants we're deciding on. We'll just honour whatever
    you choose with `tz_ambiguity_resolution`.

    :param dt datetime.datetime: A UO-aware datetime that represents the
        datetime from which we'll start looking back for `t`. `dt.tzinfo` must
        be an instance of pytz.tzinfo.BaseTzInfo and must represent an UO, not
        just a TZ.
    :param t datetime.time: The TZ-aware time we want to find out. `t` must be
        an instance of pytz.tzinfo.BaseTzInfo and represents a TZ, not an UO.
    :param tz_ambiguity_resolution: How to disambiguate the resulting time.
    :rtype: A UO-aware datetime object normalized as observable in `dt.tzinfo`.
    """
    if dt.tzinfo is None:
        raise ValueError(
            "dt doesn't have a declared UTC offset. Refusing to operate on it",
        )
    if not isinstance(dt.tzinfo, PYTZ_BASE_CLASSES):
        raise ValueError("dt's tzinfo must be a pytz instance.")

    if t.tzinfo is None:
        raise ValueError(
            "t doesn't have a declared timezone. Refusing to operate on it",
        )
    if not isinstance(t.tzinfo, PYTZ_BASE_CLASSES):
        raise ValueError("t's tzinfo must be a pytz instance.")

    current_try = _force_time_into_datetime(dt, t, tz_ambiguity_resolution)
    if current_try < dt:
        return current_try

    # If t happens to be after dt on the day that dt represents, then
    # current_try will be after dt, not before.
    # Why not subtract one day to current_try? Because that jump might not make
    # the same TZ disambiguation decisions than the first jump. That would have
    # worked iff we were working with only UOs, not TZs. t only knows its TZ.
    return _force_time_into_datetime(
        dt.tzinfo.normalize(dt - datetime.timedelta(days=1)),
        t,
        tz_ambiguity_resolution,
    )
